#lang scheme

(require "structures.scm")
(require "util.scm")
(require rnrs/hashtables-6)

(provide empty-threads wait-for-threads augment-threads
         new-monitor monitor-enter monitor-leave monitor-wait monitor-notify
         )

; internal representation

(define *threads* '())

(define (wait-for-threads)
  (map thread-wait *threads*))

(define (augment-threads t)
  (set! *threads* (cons t *threads*)))

(define (empty-threads) (set! *threads* '()))

; threads

(define-struct enter-msg (obj thread))
(define-struct leave-msg (obj thread))
(define-struct wait-msg (obj cond-name thread) #:transparent)
(define-struct notify-msg (obj cond-name thread) #:transparent)

(define (scheduler-loop)
  (let loop ((msg (thread-receive)))
    ;(printf "sched loop: ~a~n" msg)
    (cond
      ((enter-msg? msg)  (handle-enter  (enter-msg-obj msg) 
                                        (enter-msg-thread msg)))
      ((leave-msg? msg)  (handle-leave  (leave-msg-obj msg)
                                        (leave-msg-thread msg)))
      ((wait-msg? msg)   (handle-wait   (wait-msg-obj msg) 
                                        (wait-msg-cond-name msg) 
                                        (wait-msg-thread msg)))
      ((notify-msg? msg) (handle-notify (notify-msg-obj msg) 
                                        (notify-msg-cond-name msg) 
                                        (notify-msg-thread msg)))
      (else (error 'scheduler-loop "not a valid message ~a~n" msg))
      )
    (loop (thread-receive))))

(define SCHEDULER (thread scheduler-loop))
(define (monitor-enter obj)    (sched (make-enter-msg obj    (current-thread))))
(define (monitor-leave obj)    (sched (make-leave-msg obj    (current-thread))))
(define (monitor-wait obj c)   (sched (make-wait-msg  obj  c (current-thread))))
(define (monitor-notify obj c) (sched (make-notify-msg obj c (current-thread))))
(define (sched msg) 
  (thread-send SCHEDULER msg) 
  ;(printf "sending message ~a~n" (current-thread))
  (thread-suspend (current-thread)))

(define-struct monitor ((free? #:mutable) (owner #:mutable) queues))
(define (new-monitor)
  (make-monitor #t #f (make-hashtable string-hash equal?)))

(define ENTER_QUEUE " enter ")

(define (handle-enter obj thread)
  (let* ((monitor     (rt-object-monitor obj))
         (queues      (monitor-queues monitor))
         (enter-queue (hashtable-ref queues ENTER_QUEUE '())))
    (if (or (monitor-free? monitor)
            (equal? (monitor-owner monitor) thread))
        ; the thread can enter
        (begin
          (set-monitor-free?! monitor #f)
          (set-monitor-owner! monitor thread)
          (thread-resume thread))
        ; the thread must wait at the end of the "waiting-to-enter" queue
        (begin
          (hashtable-set! queues 
                          ENTER_QUEUE
                          (append enter-queue (list thread)))))))

(define (handle-leave obj thread)
  (let* ((monitor     (rt-object-monitor obj))
         (queues      (monitor-queues monitor))
         (enter-queue (hashtable-ref queues ENTER_QUEUE '())))
    (monitor-check-owner monitor thread "leave") 
    ; free the monitor & resume thread
    (set-monitor-free?! monitor #t)
    (set-monitor-owner! monitor #f)
    (thread-resume thread)
    ; also resume first "waiting-to-enter" thread (if any)   
    (cond ((not (null? enter-queue))
           (let ((first-waiting (car enter-queue)))
             (hashtable-set! queues ENTER_QUEUE (cdr enter-queue))
             ; acquire the monitor
             (set-monitor-free?! monitor #f)
             (set-monitor-owner! monitor first-waiting)
             (thread-resume first-waiting))))))

(define (handle-wait obj condition-name thread)
  (let* ((monitor     (rt-object-monitor obj))
         (queues      (monitor-queues monitor))
         (enter-queue (hashtable-ref queues ENTER_QUEUE '())))
    (monitor-check-owner monitor thread "wait")
    ; free the monitor
    (set-monitor-free?! monitor #t)
    (set-monitor-owner! monitor #f)
    ; wait at the end of the "condition-name" queue
    (hashtable-set! queues condition-name 
                    (append (hashtable-ref queues condition-name '()) 
                            (list thread)))
    ;(printf "wait: [~a] ~a~n" condition-name (hashtable-ref queues condition-name #f))
    ; and resume first "waiting-to-enter" thread (if any)
    (cond ((not (null? enter-queue))
           (let ((first-waiting (car enter-queue)))
             (hashtable-set! queues ENTER_QUEUE (cdr enter-queue))
             ; acquire the monitor
             (set-monitor-free?! monitor #f)
             (set-monitor-owner! monitor first-waiting)
             (thread-resume first-waiting))))))

(define (handle-notify obj condition-name thread)
  (let* ((monitor       (rt-object-monitor obj))
         (queues        (monitor-queues monitor))
         (enter-queue   (hashtable-ref queues ENTER_QUEUE '()))
         (waiting-queue (hashtable-ref queues condition-name '())))
    (monitor-check-owner monitor thread "notify")
    ;(printf "notify: [~a] ~a~n" condition-name waiting-queue)
    (cond ((not (null? waiting-queue))
           (let ((first-waiting (car waiting-queue)))
             (hashtable-set! queues condition-name (cdr waiting-queue))
             ; maintain the thread suspended only if we are going to resume other
             ; (in that case it must wait at the *begining* of the "waiting-to-enter" queue)
             (hashtable-set! queues 
                             ENTER_QUEUE
                             (cons thread enter-queue))
             ; and also change the monitor owner
             (set-monitor-owner! monitor first-waiting)
             (thread-resume first-waiting)))
          (else
           (thread-resume thread)))))

(define (monitor-check-owner monitor thread msg)
  (cond ((or (monitor-free? monitor) 
             (not (equal? (monitor-owner monitor) thread)))
        (error 'check-owner "thread is not the owner of the monitor: ~a~n" msg))))
